<?php

namespace Drupal\monarch_hook_plugin;

use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\Extension\ModuleHandler as CoreModuleHandler;

/**
 * The new module handler service.
 */
class ModuleHandler extends CoreModuleHandler {

  use StringTranslationTrait;

  /**
   * The hook plugin manager service.
   *
   * @var \Drupal\monarch_hook_plugin\HookPluginManager
   */
  protected $hookPluginManager;

  /**
   * The messenger service.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected $messenger;

  /**
   * Constructs a ModuleHandler object.
   *
   * @param string $root
   *   The app root.
   * @param array $module_list
   *   An associative array whose keys are the names of installed modules and
   *   whose values are Extension class parameters. This is normally the
   *   %container.modules% parameter being set up by DrupalKernel.
   * @param \Drupal\Core\Cache\CacheBackendInterface $cache_backend
   *   Cache backend for storing module hook implementation information.
   * @param \Drupal\Core\Messenger\MessengerInterface $messenger
   *   The messenger service.
   *
   * @see \Drupal\Core\DrupalKernel
   * @see \Drupal\Core\CoreServiceProvider
   */
  public function __construct($root, array $module_list, CacheBackendInterface $cache_backend, MessengerInterface $messenger) {
    parent::__construct($root, $module_list, $cache_backend);
    $this->messenger = $messenger;
  }

  /**
   * Get the hook plugin manager service.
   *
   * @return \Drupal\monarch_hook_plugin\HookPluginManager
   *   The hook plugin manager service.
   */
  protected function getHookPluginManager() {
    return $this->hookPluginManager = $this->hookPluginManager ?? \Drupal::service('plugin.manager.monarch_hook_plugin.hook');
  }

  /**
   * {@inheritdoc}
   */
  public function hasImplementations(string $hook, $modules = NULL): bool {
    if ($ret = parent::hasImplementations($hook, $modules)) {
      return $ret;
    }

    $modules = (array) $modules;

    $func_name = $this->getHookPluginManager()->getFunctionName('hook_' . $hook);

    foreach ($this->getHookPluginManager()->getPlugins($func_name) as $plugin) {
      if (
        ($provider = $plugin->getPluginDefinition()['provider'] ?? NULL) &&
        (is_null($modules) || in_array($provider, $modules))
      ) {
        return TRUE;
      }
    }

    return FALSE;
  }

  /**
   * {@inheritdoc}
   */
  public function invoke($module, $hook, array $args = []) {
    $func_name = $this->getHookPluginManager()->getFunctionName('hook_' . $hook);
    $ret = [];

    foreach ($this->getHookPluginManager()->getPlugins($func_name, -INF, 0) as $plugin) {
      if (
        ($provider = $plugin->getPluginDefinition()['provider'] ?? NULL) &&
        ($provider === $module)
      ) {
        $ret[] = $plugin->{$func_name}(...$args);
      }
    }

    if (function_exists($hook_function = $module . '_' . $hook)) {
      $ret[] = $hook_function(...$args);
    }

    foreach ($this->getHookPluginManager()->getPlugins($func_name, 0, INF) as $plugin) {
      if (
        ($provider = $plugin->getPluginDefinition()['provider'] ?? NULL) &&
        ($provider === $module)
      ) {
        $ret[] = $plugin->{$func_name}(...$args);
      }
    }

    $ret = array_filter($ret, function ($value) {
      return !is_null($value);
    });

    if (count($ret) > 1) {
      $this->messenger->addWarning($this->t('Module <em>@module</em> has multiple <em>hook_@hook</em> hook invocations, but some return values were discarded. Specific module invocation via <em>invoke</em> can only return the value of one hook.', [
        '@module' => $module,
        '@hook' => $hook,
      ]));
    }

    if (count($ret) >= 1) {
      return reset($ret);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function invokeAllWith(string $hook, callable $callback): void {
    if ($hook === 'cron') {
      print('');
    }

    $func_name = $this->getHookPluginManager()->getFunctionName('hook_' . $hook);

    foreach ($this->getHookPluginManager()->getPlugins($func_name, -INF, 0) as $plugin) {
      $callback(function (...$args) use ($func_name, $plugin) {
        return $plugin->{$func_name}(...$args);
      }, 'monarch_hook_plugin');
    }

    foreach (array_keys($this->getImplementationInfo($hook)) as $module) {
      $hookInvoker = \Closure::fromCallable($module . '_' . $hook);
      $callback($hookInvoker, $module);
    }

    foreach ($this->getHookPluginManager()->getPlugins($func_name, 0, INF) as $plugin) {
      $callback(function (...$args) use ($func_name, $plugin) {
        return $plugin->{$func_name}(...$args);
      }, 'monarch_hook_plugin');
    }
  }

  /**
   * {@inheritdoc}
   */
  public function invokeAll($hook, array $args = []) {
    $return = [];
    $this->invokeAllWith($hook, function (callable $hook, string $module) use ($args, &$return) {
      $result = $hook(...$args);
      if (isset($result) && is_array($result)) {
        $return = NestedArray::mergeDeep($return, $result);
      }
      elseif (isset($result)) {
        $return[] = $result;
      }
    });
    return $return;
  }

  /**
   * {@inheritdoc}
   */
  public function alter($type, &$data, &$context1 = NULL, &$context2 = NULL) {
    $types = is_array($type) ? $type : [$type];

    foreach ($types as $hook) {
      if (is_string($hook)) {
        $func_name = $this->getHookPluginManager()->getFunctionName('hook_' . $hook . '_alter');
        foreach ($this->getHookPluginManager()->getPlugins($func_name, -INF, 0) as $plugin) {
          $plugin->{$func_name}($data, $context1, $context2);
        }
      }
    }

    parent::alter($type, $data, $context1, $context2);

    foreach ($types as $hook) {
      if (is_string($hook)) {
        $func_name = $this->getHookPluginManager()->getFunctionName('hook_' . $hook . '_alter');
        foreach ($this->getHookPluginManager()->getPlugins($func_name, 0, INF) as $plugin) {
          $plugin->{$func_name}($data, $context1, $context2);
        }
      }
    }
  }

}
